<?php

namespace Drupal\config_overlay\Plugin\ConfigFilter;

use Drupal\config_filter\Plugin\ConfigFilterBase;
use Drupal\config_overlay\Config\ExtensionStorageFactory;
use Drupal\config_overlay\Exception\ConfigOverlayMissingExtensionListException;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides an overlay configuration filter.
 *
 * This will turn the synchronization storage into an overlay of the shipped
 * configuration of all enabled extensions.
 *
 * When writing configuration this filter will skip any configuration that is
 * identical to configuration provided by an installed extension. Thus, only
 * configuration that has been added or modified will be written to the
 * synchronization directory. When reading configuration the extensions'
 * configuration is amended to the configuration in the directory, so that the
 * missing files are transparent and configuration import, for example, works as
 * expected.
 *
 * Because shipped configuration (in contrast to the active configuration) does
 * not contain a UUID or a default configuration hash, the 'uuid' and '_core'
 * keys are ignored when comparing active and shipped configuration. When
 * reading the extension configuration as part of the overlay, the respective
 * UUIDs and hashes of the active configuration are amended automatically so
 * that the configuration import does not detect any differences relative to the
 * active configuration.
 *
 * Note that due to this the configuration system will not be able to detect
 * configuration that has been deleted and recreated with the same name when
 * this filter is active. Such configuration will be detected as an update. If
 * configuration is recreated to be exactly the same as before (but for the
 * UUID) this will not be detected as a change at all if this filter is active.
 *
 * Deleting a shipped configuration entity is supported by tracking the list of
 * deleted shipped configuration in the 'config_overlay.settings' configuration
 * object, which will appear in the configuration export once any shipped
 * configuration is deleted.
 *
 * @ConfigFilter(
 *   id = "config_overlay",
 *   label = @Translation("Configuration Overlay"),
 *   weight = 150,
 * )
 *
 * @phpstan-consistent-constructor
 */
class ConfigOverlayFilter extends ConfigFilterBase implements ContainerFactoryPluginInterface {

  /**
   * The extension storage factory.
   *
   * @var \Drupal\config_overlay\Config\ExtensionStorageFactory
   */
  protected $extensionStorageFactory;

  /**
   * The active configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $activeStorage;

  /**
   * Configuration keys to ignore in the source configuration.
   *
   * In case any of these keys are present in the source configuration but not
   * in the extension configuration, the configuration will still be considered
   * equal.
   *
   * @var string[]
   */
  protected $keysToIgnore = ['_core', 'uuid'];

  /**
   * The extension storage.
   *
   * This is a chain storage that wraps an extension storage for the
   * 'config/install' directories and an extension storage for the
   * 'config/optional' directories.
   *
   * @var \Drupal\Core\Config\StorageInterface
   *
   * @see \Drupal\Core\Config\ExtensionInstallStorage
   * @see \Drupal\config_overlay\Config\ExtensionOptionalStorage
   */
  protected $extensionStorage;

  /**
   * Constructs an extension configuration filter.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\config_overlay\Config\ExtensionStorageFactory $extension_storage_factory
   *   The extension storage factory.
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   *   The active configuration storage.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, ExtensionStorageFactory $extension_storage_factory, StorageInterface $active_storage) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->extensionStorageFactory = $extension_storage_factory;
    $this->activeStorage = $active_storage;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('config_overlay.extension_storage_factory'),
      $container->get('config.storage')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function setSourceStorage(StorageInterface $source_storage) {
    parent::setSourceStorage($source_storage);

    // The extension storage depends on the source storage to read the list of
    // installed extensions so initialize it as soon as the source storage is
    // available. When this filter is created for a non-default collection of
    // the source storage, a respective collection of the extension storage will
    // be created explicitly.
    /* @see \Drupal\config_overlay\Plugin\ConfigFilter\ExtensionConfigFilter::filterCreateCollection() */
    if ($this->getSourceStorage()->getCollectionName() === StorageInterface::DEFAULT_COLLECTION) {
      // The extension install storage requires that the passed storage contains
      // the extension list in order to work properly.
      /* @see \Drupal\Core\Config\ExtensionInstallStorage::getAllFolders() */
      $storage = $this->getStorageWithExtensionList();
      $this->setExtensionStorage($this->extensionStorageFactory->create($storage));
    }
  }

  /**
   * Sets the extension install configuration storage.
   *
   * @param \Drupal\Core\Config\StorageInterface $extension_storage
   *   The extension install configuration storage.
   */
  protected function setExtensionStorage(StorageInterface $extension_storage) {
    $this->extensionStorage = $extension_storage;
  }

  /**
   * {@inheritdoc}
   */
  public function filterExists($name, $exists) {
    if ($exists) {
      return TRUE;
    }

    $exists = $this->extensionStorage->exists($name);
    if (!$exists) {
      return FALSE;
    }

    $deleted_data = $this->getFilteredStorage()->read('config_overlay.deleted');
    return !$deleted_data || !in_array($name, $deleted_data['names'], TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function filterWriteEmptyIsDelete($name) {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  public function filterDelete($name, $delete) {
    if ($name === 'config_overlay.deleted') {
      $delete = FALSE;
    }

    return $delete;
  }

  /**
   * {@inheritdoc}
   */
  public function filterRead($name, $data) {
    // If the configuration does not exist in the sync directory, look it up
    // in the extensions' configuration.
    if (!$data) {
      $data = $this->extensionStorage->read($name);

      // Amend the exported configuration with the UUID and hash of the active
      // configuration.
      if ($data && ($active_data = $this->activeStorage->read($name))) {
        $this->processExtensionDataOnRead($data, $active_data);
      }
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function filterReadMultiple(array $names, array $data) {
    // Look up any missing configuration in the extensions' configuration.
    $remaining_names = array_diff($names, array_keys($data));
    $data += $this->extensionStorage->readMultiple($remaining_names);

    // Amend the exported configuration with the UUID and hash of the active
    // configuration.
    $active_data = $this->activeStorage->readMultiple($remaining_names);
    foreach ($remaining_names as $remaining_name) {
      if (isset($data[$remaining_name]) && isset($active_data[$remaining_name])) {
        $this->processExtensionDataOnRead($data[$remaining_name], $active_data[$remaining_name]);
      }
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function filterWrite($name, array $data) {
    $extension_data = $this->extensionStorage->read($name);

    // @todo Fix this
    if (($name === 'config_overlay.deleted') && !$extension_data) {
      $extension_data = ['names' => []];
    }

    if ($extension_data) {
      $data_to_compare = $data;
      foreach ($this->keysToIgnore as $key_to_ignore) {
        if (!isset($extension_data[$key_to_ignore])) {
          unset($data_to_compare[$key_to_ignore]);
        }
      }

      // Do not write anything if the active configuration matches configuration
      // provided by an extension.
      if ($data_to_compare === $extension_data) {
        return;
      }
    }

    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function filterListAll($prefix, array $data) {
    $extension_data = array_diff($this->extensionStorage->listAll($prefix), $data);
    $data = array_filter($data, [$this->getFilteredStorage(), 'exists']);
    return array_merge($data, $extension_data);
  }

  /**
   * {@inheritdoc}
   */
  public function filterDeleteAll($prefix, $delete) {
    if ($delete) {
      $deleted_data = $this->getFilteredStorage()->read('config_overlay.deleted');
      // When configuration is being exported for the first time after Config
      // Overlay has been installed, the deletion configuration is not yet
      // present.
      if ($deleted_data && $deleted_data['names']) {
        // Deleting all configuration would delete the deletion information
        // itself, so the configuration must be deleted individually.
        $delete = FALSE;
      }
    }

    return $delete;
  }

  /**
   * {@inheritdoc}
   */
  public function filterCreateCollection($collection) {
    $filter = new static(
      $this->configuration,
      $this->pluginId,
      $this->pluginDefinition,
      $this->extensionStorageFactory,
      $this->activeStorage->createCollection($collection)
    );
    $filter->setExtensionStorage($this->extensionStorage->createCollection($collection));
    return $filter;
  }

  /**
   * Adds ignored data from the active configuration to the exported data.
   *
   * @param array $extension_data
   *   The respective extension configuration data that is being read.
   * @param array $active_data
   *   The respective active configuration data.
   */
  public function processExtensionDataOnRead(array &$extension_data, array $active_data) {
    foreach ($this->keysToIgnore as $key_to_ignore) {
      // The system.site configuration specifies an empty UUID, so checking
      // whether the 'uuid' key is set is not sufficient.
      if (empty($extension_data[$key_to_ignore]) && isset($active_data[$key_to_ignore])) {
        $extension_data[$key_to_ignore] = $active_data[$key_to_ignore];
      }
    }

    // Make sure that the ignored data is positioned in the same place in the
    // data array as it is in the active configuration so that strict equality
    // between the exported and active configuration can be achieved. The
    // intersection makes sure that other keys that are available in the active
    // configuration but not in the exported configuration are not merged.
    $extension_data = array_intersect_key(array_merge($active_data, $extension_data), $extension_data);
  }

  /**
   * Gets the storage with the extension list.
   *
   * Generally the extension list is read from the source storage, so that
   * configuration will be found in modules that will be installed by importing
   * configuration. Prior to the very first configuration export, however, the
   * source storage is empty, and so the 'core.extension' configuration does not
   * exist either. In this case the extension list is read from the active
   * storage, so that this filter can work at all.
   *
   * @return \Drupal\Core\Config\StorageInterface
   *   The storage that contains the extension list.
   *
   * @throws \Drupal\config_overlay\Exception\ConfigOverlayMissingExtensionListException
   *   If no storage with an extension could be found.
   */
  protected function getStorageWithExtensionList() {
    foreach ([$this->getSourceStorage(), $this->activeStorage] as $possible_storage) {
      if ($possible_storage->exists('core.extension')) {
        return $possible_storage;
      }
    }

    throw new ConfigOverlayMissingExtensionListException('The extension list could not be found in the source or the active storage');
  }

}
